/*
 * Hanashizume no Matsuri for X
 * Copyright (c) 2000 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: image.c,v 1.13 2000/11/27 13:09:18 tf Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "hana.h"

static struct {
    int id;
    TrueColorMap *tcmap;
} tcmap[10];

static int tcmap_idx = 0;

static int a_map[17][512]; /* alpha factor */
static int i_map[33][256]; /* intensity factor */

static void get_visual_info(Display *, Window, Image *, XWindowAttributes);

Image *ImageCreate(Display *d, Window w, int width, int height)
{
  int major, minor;
  Image *image;
  int depth;
  Visual *visual;
  XWindowAttributes att;
  
  if ((image = (Image *)malloc(sizeof(Image))) == NULL) {
    perror("ImageCreate:malloc");
    exit(1);
  }

  image->display = d;
  image->window  = w;
  image->xoffset = 0;
  image->yoffset = 0;
  image->rgbmap = NULL;
  image->pixel_num = 0;

  if ((image->support_shmem = XShmQueryExtension(d))) {
    XShmQueryVersion(d, &major, &minor, &image->support_shpix);
    fprintf(stderr, "Support MIT-Shmem Extention %d.%d\n", major, minor);
  }
  /* image->support_shmem = False; */

  XGetWindowAttributes(d, w, &att);
  depth = att.depth;
  visual = att.visual;

  image->colormap = att.colormap;
  image->gc = DefaultGC(d, XScreenNumberOfScreen(att.screen));

  /* get visual information */
  get_visual_info(d, w, image, att);

  if (image->support_shmem) {
    fprintf(stderr, "Using shared images.\n");

    image->ximage = XShmCreateImage(d, visual, depth, ZPixmap, NULL,
				    &image->shminfo, width, height);

    if ((image->shminfo.shmid = 
	 shmget(IPC_PRIVATE, image->ximage->bytes_per_line * height,
		IPC_CREAT | 0777)) < 0) {
      fprintf(stderr, "shmget: size:%x %s\n",
	      image->ximage->bytes_per_line * height,
	      strerror(errno));
      exit(1);
    }
    
    if ((image->shminfo.shmaddr = image->ximage->data = 
	 shmat(image->shminfo.shmid, 0, 0)) == NULL) {
      perror("shmat");
      exit(1);
    }

    image->shminfo.readOnly = False;
    XShmAttach(d, &image->shminfo);

    /* shared pixmap */
    if (image->support_shpix) {
      fprintf(stderr, "Get shared pixmap\n");
      
      image->pixmap = XShmCreatePixmap(d, w,
				       image->ximage->data, 
				       &image->shminfo,
				       image->ximage->width,
				       image->ximage->height, 
				       image->ximage->depth);
    } else {
      fprintf(stderr, "Get normal pixmap\n");
      image->pixmap = XCreatePixmap(d, w, 
				    image->ximage->width,
				    image->ximage->height,
				    image->ximage->depth);
    }
  } else {
    fprintf(stderr, "Get normal image\n");
    image->ximage = XCreateImage(d, visual, depth, ZPixmap, 0, NULL,
				 width, height, 8, 0);
    
    if((image->ximage->data = malloc(image->ximage->bytes_per_line *
				     image->ximage->height)) == NULL) {
      perror("ImageCreate");
      exit(1);
    }
    image->pixmap = XCreatePixmap(d, w, 
				  image->ximage->width,
				  image->ximage->height,
				  image->ximage->depth);
  }
  
  return image;
}

static void get_visual_info(Display *d, Window w, Image *image, XWindowAttributes att)
{
  int matched;
  XVisualInfo *vinfolist;
  XVisualInfo vinfo;

  /* get visual */
  vinfo.visualid = XVisualIDFromVisual(att.visual);
  vinfolist = XGetVisualInfo(d, VisualIDMask, &vinfo, &matched);
  if (!matched) {
    fprintf(stderr, "Can't get visual info.\n");
    exit(1);
  }
  image->visual_class = vinfolist[0].class;
  XFree(vinfolist);

  if (image->visual_class == TrueColor || image->visual_class == DirectColor) {
    int i;
    TrueColorMap *map;

    fprintf(stderr, "TrueColor mode\n");
    
    for (i = 0; i < tcmap_idx; i++) {
      if (vinfo.visualid == tcmap[i].id) {
	image->tcmap = tcmap[i].tcmap;
	goto end;
      }
    }
    if ((map = (TrueColorMap *)malloc(sizeof(TrueColorMap))) == NULL) {
      perror("get_visual_info: malloc");
      exit(1);
    }
    /* Get colormap for TrueColor */
    for (i = 0; i < 256; i++) {
      int col;
      XColor defs;
      defs.flags = DoRed | DoGreen | DoBlue;
      
      defs.red   = col = (i << 8) | i;
      defs.green = 0;
      defs.blue  = 0;
      XAllocColor(d, image->colormap, &defs);
      map->red[i] = defs.pixel;
      
      defs.red   = 0;
      defs.green = col;
      defs.blue  = 0;
      XAllocColor(d, image->colormap, &defs);
      map->green[i] = defs.pixel;
      
      defs.red   = 0;
      defs.green = 0;
      defs.blue  = col;
      XAllocColor(d, image->colormap, &defs);
      map->blue[i] = defs.pixel;
    }
    image->tcmap = tcmap[tcmap_idx++].tcmap = map;
  }
end:
}

void ImageDestroy(Image *image)
{
  if (image == NULL)
    return;

  if (image->pixel_num > 0) {
    XFreeColors(image->display, image->colormap,
		image->allocated_pixel, image->pixel_num, 0);
    image->pixel_num = 0;
  }

  XFreePixmap(image->display, image->pixmap);
  if (image->support_shmem) {
    /* detach shared memory */
    XShmDetach(image->display, &image->shminfo);
    image->ximage->data = NULL;
    /* delete shared memory */
    shmdt(image->shminfo.shmaddr);
    shmctl(image->shminfo.shmid, IPC_RMID, NULL);
  }

  /* destroy image */
  XDestroyImage(image->ximage);

  if (image->rgbmap != NULL) {
    free(image->rgbmap);
  }
  
  free(image);
}

void ImagePut(Image *image)
{
  if (image->support_shmem) {
    if (!image->support_shpix) {
      /* no shared pixmap */
      XShmPutImage(image->display, image->pixmap, image->gc, 
		   image->ximage, 0, 0, 0, 0, 
		   image->ximage->width, 
		   image->ximage->height, False);
    } else {
      /* do nothing, because shared memory works! */
    }
  } else {
    XPutImage(image->display, image->pixmap, image->gc, 
	      image->ximage, 0, 0, 0, 0, 
	      image->ximage->width,
	      image->ximage->height);
  }
}

/*
 * interface functions
 */
#define SXIMAGE_TPIXEL_SCREEN(s,r,g,b,p)\
(s->tcmap->red[i_map[p][r]]|s->tcmap->green[i_map[p][g]]|s->tcmap->blue[i_map[p][b]])

void set_ximage(Image *image, int width, int height, int density, GRAYSCALE grayscale)
{
  int i, j, idx;
  long pixel;
  int r, g, b;
  int gray;

  if (density > MAX_DENSITY) {
    fprintf(stderr, "set_ximage: Density[%d] is too large\n", density);
    density = MAX_DENSITY;
  }
  switch (image->visual_class) {
  case TrueColor:
  case DirectColor:
    for (i = 0; i < height; i++) {
      for (j = 0; j < width; j++) {
	idx = (j  + width * i) * 3;
	r = image->rgbmap[idx];
	g = image->rgbmap[idx + 1];
	b = image->rgbmap[idx + 2];
	if (grayscale.enable) {
	  gray = r * 77 + g * 150 + b * 29;
	  r = gray * grayscale.r >> 16;
	  g = gray * grayscale.g >> 16;
	  b = gray * grayscale.b >> 16;
	}
	pixel = SXIMAGE_TPIXEL_SCREEN(image, r, g, b, density);
	XPutPixel(image->ximage, j, i, pixel);
      }
    }
    break;
  default:
    fprintf(stderr, "set_ximage: Visual not supported.\n");
    exit(1);
    break;
  }
}

void mix_ximages(Image *base, Image *add, int width, int height, int ratio, int density, GRAYSCALE grayscale)
{
  int i, j, p;
  long pixel;
  unsigned char r, g, b;
  int gray;

  if (base->visual_class != add->visual_class) {
    fprintf(stderr, "Two images should have same visual classes.\n");
    exit(1);
  }

  switch (base->visual_class) {
  case TrueColor:
  case DirectColor:
    if (grayscale.enable) {
      for (i = 0; i < height; i++) {
	for (j = 0; j < width; j++) {
	  p = (j + width * i) * 3;
	  r = base->rgbmap[p] + 
	    a_map[ratio][add->rgbmap[p] - base->rgbmap[p] + 255];
	  g = base->rgbmap[p + 1] +
	    a_map[ratio][add->rgbmap[p + 1] - base->rgbmap[p + 1] + 255];
	  b = base->rgbmap[p + 2] +
	    a_map[ratio][add->rgbmap[p + 2] - base->rgbmap[p + 2] + 255];

	  gray = r * 77 + g * 150 + b * 29;
	  r = gray * grayscale.r >> 16;
	  g = gray * grayscale.g >> 16;
	  b = gray * grayscale.b >> 16;
	  
	  pixel = SXIMAGE_TPIXEL_SCREEN(base, r, g, b, density);
	  XPutPixel(base->ximage, j, i, pixel);
	}
      }
    } else {
      for (i = 0; i < height; i++) {
	for (j = 0; j < width; j++) {
	  p = (j + width * i) * 3;
	  r = base->rgbmap[p] + 
	    a_map[ratio][add->rgbmap[p] - base->rgbmap[p] + 255];
	  g = base->rgbmap[p + 1] +
	    a_map[ratio][add->rgbmap[p + 1] - base->rgbmap[p + 1] + 255];
	  b = base->rgbmap[p + 2] +
	    a_map[ratio][add->rgbmap[p + 2] - base->rgbmap[p + 2] + 255];
	  pixel = SXIMAGE_TPIXEL_SCREEN(base, r, g, b, density);
	  XPutPixel(base->ximage, j, i, pixel);
	}
      }
    }
    break;
  default:
    fprintf(stderr, "mix_ximages: Visual not supported\n");
    exit(1);
    break;
  }
}

void ImageAttributeInit(ImageAttribute *attr)
{
  GRAYSCALE g = {False, 256, 256, 256};

  attr->grayscale = g;
  attr->fade = False;
  attr->fadetime = 0;
}

void SetWhite(Image *image, int width, int height)
{
  int i;
  GRAYSCALE g = {False, 0, 0, 0};

  image->rgbmap = (unsigned char *)calloc(width * height * 3, sizeof(unsigned char));
  if (image->rgbmap == NULL) {
    perror("SetWhite: calloc");
    exit(1);
  }
  for (i = 0; i < width * height * 3; i++) {
    image->rgbmap[i] = 0xff;
  }
  set_ximage(image, width, height, MAX_DENSITY, g);
  ImagePut(image);
}

void InitImageProcTable()
{
  int i, j;

  /* alpha map */
  for (i = 0; i < 17; i++) {
    for (j = -255; j < 256; j++) {
      a_map[i][j + 255] = i * j / 16;
    }
  }
  /* intensity map */
  for (i = 0; i < 33; i++) {
    for (j = 0; j < 256; j++) {
      i_map[i][j] = i * j / 32;
    }
  }
}

#if 0
/* main procedure for test */
int main(int argc, char *argv[])
{
  Image *image;
  Display *d;
  Window w;
  int width, height;
  int i;

  width = 640;
  height = 480;
 
  d = XOpenDisplay (NULL);
  w = XCreateSimpleWindow (d, RootWindow(d, 0), 400, 400, width, height, 2, 0, 0);
  XMapWindow(d, w);
  
  image = ImageCreate(d, w, width, height);

  for (i = 255; i > 0; i-=4) {
    put_image(image, width, height, i);
    ImagePut(image);
    XCopyArea(d, image->pixmap, w, image->gc, 0, 0, width, height, 0, 0);
    XSync(d, False);
  }
  /* getchar(); */

  ImageDestroy(image);

  return 0;
}
#endif
